import datetime
from typing import Type, Dict, Set, Any, Callable, Tuple, Optional

import dacite

from dagshub.common.util import wrap_bytes
from dagshub.data_engine.client.models import IntegrationStatus, DatasourceType, PreprocessingStatus
from dagshub.data_engine.dtypes import MetadataFieldType
from dagshub.data_engine.annotation import MetadataAnnotations

metadata_type_lookup = {
    int: MetadataFieldType.INTEGER,
    bool: MetadataFieldType.BOOLEAN,
    float: MetadataFieldType.FLOAT,
    str: MetadataFieldType.STRING,
    bytes: MetadataFieldType.BLOB,
    datetime.datetime: MetadataFieldType.DATETIME,
}


metadata_type_lookup_reverse: Dict[str, Type] = {}
for k, v in metadata_type_lookup.items():
    metadata_type_lookup_reverse[v.value] = k


def default_metadata_type_value(metadata_type: MetadataFieldType) -> Any:
    if metadata_type == MetadataFieldType.DATETIME:
        return datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
    else:
        return metadata_type_lookup_reverse[metadata_type.value]()


def timestamp_to_datetime(timestamp: int) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)


dacite_config = dacite.Config(
    cast=[IntegrationStatus, DatasourceType, PreprocessingStatus, MetadataFieldType, Set],
    type_hooks={datetime.datetime: timestamp_to_datetime},
)


# Special handling of metadata conversion for certain classes
def _convert_annotation_metadata(annotation: MetadataAnnotations) -> Tuple[Optional[str], MetadataFieldType]:
    blob_value = annotation.to_ls_task()
    if blob_value is None:
        return None, MetadataFieldType.BLOB
    return wrap_bytes(blob_value), MetadataFieldType.BLOB


special_metadata_handlers: Dict[Type[Any], Callable[[Any], Tuple[Optional[str], MetadataFieldType]]] = {
    MetadataAnnotations: _convert_annotation_metadata,
}



import datetime
import inspect
import logging
from collections import Counter, defaultdict
from concurrent.futures import as_completed, ThreadPoolExecutor
from dataclasses import field, dataclass
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, List, Dict, Any, Optional, Union, Tuple, Literal, Callable
import json
import os
import os.path
import importlib

import dacite
import dagshub_annotation_converter.converters.yolo
import rich.progress
from dagshub_annotation_converter.formats.yolo import YoloContext
from dagshub_annotation_converter.formats.yolo.categories import Categories
from dagshub_annotation_converter.formats.yolo.common import ir_mapping
from dagshub_annotation_converter.ir.image import IRImageAnnotationBase
from pydantic import ValidationError

from dagshub.auth import get_token
from dagshub.common import config
from dagshub.common.analytics import send_analytics_event
from dagshub.common.api import UserAPI
from dagshub.common.download import download_files
from dagshub.common.helpers import sizeof_fmt, prompt_user, log_message
from dagshub.common.rich_util import get_rich_progress
from dagshub.common.util import lazy_load, multi_urljoin
from dagshub.data_engine.annotation import MetadataAnnotations
from dagshub.data_engine.annotation.voxel_conversion import (
    add_voxel_annotations,
    add_ls_annotations,
)
from dagshub.data_engine.client.models import DatasourceType, MetadataSelectFieldSchema
from dagshub.data_engine.model.datapoint import Datapoint, _get_blob, _generated_fields
from dagshub.data_engine.client.loaders.base import DagsHubDataset
from dagshub.data_engine.model.schema_util import dacite_config
from dagshub.data_engine.voxel_plugin_server.utils import set_voxel_envvars
from dagshub.data_engine.dtypes import MetadataFieldType

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource
    import fiftyone as fo
    import dagshub.data_engine.voxel_plugin_server.server as plugin_server_module
    import datasets as hf_ds
    import tensorflow as tf
    import mlflow
    import mlflow.entities
else:
    plugin_server_module = lazy_load("dagshub.data_engine.voxel_plugin_server.server")
    fo = lazy_load("fiftyone")
    tf = lazy_load("tensorflow")
    hf_ds = lazy_load("datasets")
    mlflow = lazy_load("mlflow")

logger = logging.getLogger(__name__)

CustomPredictor = Callable[
    [
        List[str],
    ],
    List[Tuple[Any, Optional[float]]],
]


class VisualizeError(Exception):
    """:meta private:"""

    pass


def identity_func(x):
    return x


@dataclass
class QueryResult:
    """
    Result of executing a query on a :class:`.Datasource`.

    You can iterate over this object to get the :class:`datapoints <.Datapoint>`::

        res = ds.head()
        for dp in res:
            print(dp.path_in_repo)

    """

    _entries: List[Datapoint]
    datasource: "Datasource"
    fields: List[MetadataSelectFieldSchema]
    query_data_time: Optional[datetime.datetime] = None
    _datapoint_path_lookup: Dict[str, Datapoint] = field(init=False)

    def __post_init__(self):
        self._refresh_lookups()

    @property
    def entries(self):
        """
        list(Datapoint): Datapoints contained in this QueryResult
        """
        return self._entries

    @entries.setter
    def entries(self, value: List[Datapoint]):
        self._entries = value
        self._refresh_lookups()

    def _refresh_lookups(self):
        self._datapoint_path_lookup = {}
        for e in self.entries:
            self._datapoint_path_lookup[e.path] = e

    @property
    def dataframe(self):
        """
        Represent the contents of this QueryResult as a
        `pandas.DataFrame <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html>`_.

        The created dataframe has a copy of the QueryResult's data.
        """
        import pandas as pd

        metadata_key_set = set()
        for e in self.entries:
            metadata_key_set.update(e.metadata.keys())

        metadata_keys = list(sorted(metadata_key_set))
        return pd.DataFrame.from_records([dp.to_dict(metadata_keys) for dp in self.entries])

    def __len__(self):
        return len(self.entries)

    def __iter__(self):
        """You can iterate over a QueryResult to get containing datapoints"""
        return self.entries.__iter__()

    def __repr__(self):
        return f"QueryResult of datasource {self.datasource.source.name} with {len(self.entries)} datapoint(s)"

    @staticmethod
    def from_gql_query(query_resp: Dict[str, Any], datasource: "Datasource") -> "QueryResult":
        raw_fields = query_resp.get("selectFields") or []
        fields = [dacite.from_dict(MetadataSelectFieldSchema, f, dacite_config) for f in raw_fields]
        # If no fields - get the default datasource ones
        if len(fields) == 0:
            fields = [MetadataSelectFieldSchema.from_metadata_field_schema(mfs) for mfs in datasource.fields]

        edges = query_resp.get("edges", [])
        if edges is None:
            edges = []
        datapoints = [Datapoint.from_gql_edge(edge, datasource, fields) for edge in edges]
        query_data_time = datetime.datetime.fromtimestamp(query_resp.get("queryDataTime"), tz=datetime.timezone.utc)

        return QueryResult(_entries=datapoints, datasource=datasource, fields=fields, query_data_time=query_data_time)

    def as_ml_dataset(self, flavor: str, **kwargs):
        """
        Convert the QueryResult into a dataset for a machine learning framework

        Args:
            flavor: Either of:

                - ``"torch"``: returns a \
                `torch.utils.data.Dataset <https://pytorch.org/docs/stable/data.html#torch.utils.data.Dataset>`_
                - ``"tensorflow"``: returns a \
                `tf.data.Dataset <https://www.tensorflow.org/api_docs/python/tf/data/Dataset>`_


        Keyword Args:
            metadata_columns (list(str)): which fields to use in the dataset.
            strategy (str): Datapoint file loading strategy. Possible values:

                - ``"preload"`` - Download all datapoints before returning the dataset.
                - ``"background"`` - Start downloading the datapoints in the background. \
                    If an undownloaded datapoint is accessed, it gets downloaded.
                - ``"lazy"`` (default) - Download each datapoint as it is being accessed by the dataset.

            savedir (str|Path): Where to store the datapoint files. Default is :func:`datasource's default location \
                <dagshub.data_engine.model.datasource.Datasource.default_dataset_location>`
            processes (int): number of parallel processes to download the datapoints with. Default is 8.
            tensorizers: How to transform the datapoint file/metadata into tensors. Possible values:

                - ``"auto"`` - try to guess the tensorizers for every field.\
                    For files the tensorizer will be the determined by the first file's extension.
                - ``"image"`` | ``"audio"`` - tensorize all fields according to this type
                - List of tensorizers. First is the tensorizer for the datapoint file and then \
                    a tensorizer for each of the metadata fields. Tensorizer can either be strings\
                     ``"image"``, ``"audio"``, ``"numeric"``, or your own function that receives the\
                     metadata value of the field and turns it into a tensor.
        """
        flavor = flavor.lower()
        if flavor == "torch":
            from dagshub.data_engine.client.loaders.torch import PyTorchDataset

            return PyTorchDataset(self, **kwargs)
        elif flavor == "tensorflow":
            from dagshub.data_engine.client.loaders.tf import TensorFlowDataset

            ds_builder = TensorFlowDataset(self, **kwargs)
            ds = tf.data.Dataset.from_generator(ds_builder.generator, output_signature=ds_builder.signature)
            ds.__len__ = lambda: ds_builder.__len__()
            ds.__getitem__ = ds_builder.__getitem__
            ds.builder = ds_builder
            ds.type = "tensorflow"
            return ds
        else:
            raise ValueError("supported flavors are torch|tensorflow")

    def as_ml_dataloader(self, flavor, **kwargs):
        """
        Convert the QueryResult into a dataloader for a machine learning framework

        Args:
            flavor: Either of:

                - ``"torch"``: returns a \
                `torch.utils.data.DataLoader <https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader>`_
                - ``"tensorflow"``: returns a \
                `tf.keras.utils.Sequence <https://www.tensorflow.org/api_docs/python/tf/keras/utils/Sequence>`_

        Kwargs are the same as :func:`as_ml_dataset()`.
        """

        send_analytics_event(
            "Client_DataEngine_DataLoaderInitialized",
            repo=self.datasource.source.repoApi,
        )

        def keypairs(keys):
            return {key: kwargs[key] for key in keys}

        if type(flavor) is not str:
            if flavor.type == "torch":
                from dagshub.data_engine.client.loaders.torch import PyTorchDataLoader

                return PyTorchDataLoader(flavor, **kwargs)
            elif flavor.type == "tensorflow":
                from dagshub.data_engine.client.loaders.tf import TensorFlowDataLoader

                return TensorFlowDataLoader(flavor, **kwargs)

        kwargs["for_dataloader"] = True
        dataset_kwargs = set(list(inspect.signature(DagsHubDataset).parameters.keys())[1:])
        global_kwargs = set(kwargs.keys())
        flavor = flavor.lower() if type(flavor) is str else flavor
        if flavor == "torch":
            from dagshub.data_engine.client.loaders.torch import PyTorchDataLoader

            return PyTorchDataLoader(
                self.as_ml_dataset(flavor, **keypairs(global_kwargs.intersection(dataset_kwargs))),
                **keypairs(global_kwargs - dataset_kwargs),
            )
        elif flavor == "tensorflow":
            from dagshub.data_engine.client.loaders.tf import TensorFlowDataLoader

            return TensorFlowDataLoader(
                self.as_ml_dataset(
                    flavor,
                    **keypairs(global_kwargs.intersection(dataset_kwargs)),
                ),
                **keypairs(global_kwargs - dataset_kwargs),
            )
        else:
            raise ValueError("supported flavors are torch|tensorflow|<torch.utils.data.Dataset>|<tf.data.Dataset>")

    def as_hf_dataset(
        self, target_dir: Optional[Union[str, PathLike]] = None, download_datapoints=True, download_blobs=True
    ):
        """
        Loads this QueryResult as a HuggingFace dataset.

        The paths of the downloads are set to the local paths in the filesystem, so they can be used with
        a `cast_column() \
        <https://huggingface.co/docs/datasets/main/en/package_reference/main_classes#datasets.Dataset.cast_column>`_\
        function later.

        Args:
            target_dir: Where to download the datapoints. The metadata is still downloaded into the global cache.
            download_datapoints: If set to ``True`` (default), downloads the datapoint files and sets the path column\
                to the path of the datapoint in the filesystem
            download_blobs: If set to ``True`` (default), downloads all blob fields and sets the respective column\
                to the path of the file in the filesystem.
        """
        if download_blobs:
            # Download blobs as paths, so later a user can apply ds.cast_column on the blobs
            self.get_blob_fields(load_into_memory=False, path_format="str")

        df = self.dataframe

        if download_datapoints:
            # Do the same for the actual datapoint files, changing the path

            if target_dir is None:
                target_dir = self.datasource.default_dataset_location
            elif isinstance(target_dir, str):
                target_dir = Path(target_dir).absolute()
            new_paths = []
            self.download_files(target_dir=target_dir)
            for dp in df["path"]:
                new_paths.append(str(target_dir / self.datasource.source.source_prefix / dp))
            df["path"] = new_paths

        # Drop the generated fields
        for f in _generated_fields.keys():
            if f == "path":
                continue
            df.drop(f, axis=1, inplace=True)

        return hf_ds.Dataset.from_pandas(df)

    def __getitem__(self, item: Union[str, int, slice]):
        """
        Gets datapoint by its path (string) or by its index in the result (or slice)
        """
        if isinstance(item, str):
            return self._datapoint_path_lookup[item]
        elif isinstance(item, int):
            return self.entries[item]
        elif type(item) is slice:
            return QueryResult(_entries=self.entries[item], datasource=self.datasource, fields=self.fields)
        else:
            raise ValueError(
                f"Can't lookup datapoint using value {item} of type {type(item)}, "
                f"needs to be either int or str or slice"
            )

    def get_blob_fields(
        self,
        *fields: str,
        load_into_memory=False,
        cache_on_disk=True,
        num_proc: int = config.download_threads,
        path_format: Literal["str", "path"] = "path",
    ) -> "QueryResult":
        """
        Downloads data from blob fields.

        If ``load_into_memory`` is set to ``True``, then we additionally convert special fields to new types:

            - Annotation fields are converted to \
                :class:`MetadataAnnotations <dagshub.data_engine.annotation.metadata.MetadataAnnotations>`
            - Document fields are converted to strings

        Args:
            fields: list of binary fields to download blobs for. If empty, download all blob fields.
            load_into_memory: Whether to load the blobs into the datapoints, or just store them on disk


                If True: the datapoints' specified fields will contain the blob data

                If False: the datapoints' specified fields will contain Path objects to the file of the downloaded blob
            cache_on_disk: Whether to cache the blobs on disk or not (valid only if load_into_memory is set to True)
                Cache location is ``~/dagshub/datasets/<repo>/<datasource_id>/.metadata_blobs/``
            num_proc: number of download threads
            path_format: What way the paths to the file should be represented.
                ``path`` returns a Path object, and ``str`` returns a string of this path.
        """
        send_analytics_event("Client_DataEngine_downloadBlobs", repo=self.datasource.source.repoApi)
        if not load_into_memory:
            assert cache_on_disk

        # If no fields are specified, include all blob fields from self..fields
        if not fields:
            fields = tuple(
                [
                    f.name
                    for f in self.fields
                    if f.valueType == MetadataFieldType.BLOB and f.name not in self.document_fields
                ]
            )

        if len(fields) == 0:
            logger.warning("No blob fields loaded")
            return self

        cache_location = self.datasource.default_dataset_location / ".metadata_blobs"

        if cache_on_disk:
            cache_location.mkdir(parents=True, exist_ok=True)

        # Create a list of things to download: (datapoint, field, url, path_location)
        to_download: List[Tuple[Datapoint, str, str, Path]] = []
        for dp in self.entries:
            for fld in fields:
                field_value = dp.metadata.get(fld)
                # If field_value is a blob or a path, then ignore, means it's already been downloaded
                if not isinstance(field_value, str):
                    continue
                download_task = (dp, fld, dp.blob_url(field_value), dp.blob_cache_location / field_value)
                to_download.append(download_task)

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())
        task = progress.add_task("Downloading blobs...", total=len(to_download))

        auth = self.datasource.source.repoApi.auth

        def _get_blob_fn(dp: Datapoint, field: str, url: str, blob_path: Path):
            blob_or_path = _get_blob(url, blob_path, auth, cache_on_disk, load_into_memory, path_format)
            if isinstance(blob_or_path, str) and path_format != "str":
                logger.warning(f"Error while downloading blob for field {field} in datapoint {dp.path}:{blob_or_path}")
            dp.metadata[field] = blob_or_path

        with progress:
            with ThreadPoolExecutor(max_workers=num_proc) as tp:
                futures = [tp.submit(_get_blob_fn, *args) for args in to_download]
                for f in as_completed(futures):
                    exc = f.exception()
                    if exc is not None:
                        logger.warning(f"Got exception {type(exc)} while downloading blob: {exc}")
                    progress.update(task, advance=1)

        self._convert_annotation_fields(*fields, load_into_memory=load_into_memory)

        # Convert any downloaded document fields
        document_fields = [f for f in fields if f in self.document_fields]
        if document_fields:
            for dp in self:
                for fld in document_fields:
                    if fld in dp.metadata:
                        # Override the load_into_memory flag, because we need the contents
                        if not load_into_memory:
                            dp.metadata[fld] = Path(dp.metadata[fld]).read_bytes()
                        dp.metadata[fld] = dp.metadata[fld].decode("utf-8")

        return self

    def _convert_annotation_fields(self, *fields, load_into_memory):
        # Convert any downloaded annotation column
        annotation_fields = [f for f in fields if f in self.annotation_fields]

        bad_annotations = defaultdict(list)

        if annotation_fields:
            # Convert them
            for dp in self:
                for fld in annotation_fields:
                    if fld in dp.metadata:
                        # Already loaded - skip
                        if isinstance(dp.metadata[fld], MetadataAnnotations):
                            continue
                        # Override the load_into_memory flag, because we need the contents
                        if not load_into_memory:
                            dp.metadata[fld] = Path(dp.metadata[fld]).read_bytes()
                        try:
                            dp.metadata[fld] = MetadataAnnotations.from_ls_task(
                                datapoint=dp, field=fld, ls_task=dp.metadata[fld]
                            )
                        except ValidationError:
                            bad_annotations[fld].append(dp.path)
                    else:
                        dp.metadata[fld] = MetadataAnnotations(datapoint=dp, field=fld)

        if bad_annotations:
            log_message(
                "Warning: The following datapoints had invalid annotations, "
                "any annotation-related operations will not work on these:"
            )
            err_msg = ""
            for fld, dps in bad_annotations.items():
                err_msg += f'Field "{fld}" in datapoints:\n\t'
                err_msg += "\n\t".join(dps)
            log_message(err_msg)

    def download_binary_columns(
        self,
        *columns: str,
        load_into_memory=True,
        cache_on_disk=True,
        num_proc: int = 32,
    ) -> "QueryResult":
        """
        deprecated: Use get_blob_fields instead.

        :meta private:
        """
        return self.get_blob_fields(
            *columns,
            load_into_memory=load_into_memory,
            cache_on_disk=cache_on_disk,
            num_proc=num_proc,
        )

    def annotate_with_mlflow_model(
        self,
        repo: str,
        name: str,
        host: Optional[str] = None,
        version: str = "latest",
        pre_hook: Callable[[List[str]], Any] = identity_func,
        post_hook: Callable[[Any], Any] = identity_func,
        batch_size: int = 1,
        log_to_field: str = "annotation",
    ) -> Dict[str, Any]:
        """
        Fetch an MLflow model from a specific repository and use it to annotate the datapoints in this QueryResult.
        The resulting annotations are then stored in the field specified by ``log_to_field``.

        Any MLflow model that has a ``model.predict`` endpoint is supported.
        This includes, but is not limited to the following flavors:

        * ``torch``
        * ``tensorflow``
        * ``pyfunc``
        * ``scikit-learn``

        Keep in mind that by default ``mlflow.predict()`` will receive the list of downloaded datapoint paths as input,
        so additional "massaging" of the data might be required for prediction to work.
        Use the ``pre_hook`` function to do so.

        Args:
            repo: repository to extract the model from
            name: name of the model in the repository's MLflow registry.
            host: address of the DagsHub instance with the repo to load the model from.
                 Set it if the model is hosted on a different DagsHub instance than the datasource.
            version: version of the model in the mlflow registry.
            pre_hook: function that runs before datapoints are sent to ``model.predict()``.
                The input argument is the list of paths to datapoint files in the current batch.
            post_hook: function that converts the model output to the desired format.
            batch_size: Size of the file batches that are sent to ``model.predict()``.
                Default batch size is 1, but it is still being sent as a list for consistency.
            log_to_field: Field to store the resulting annotations in.
        """
        res = self.predict_with_mlflow_model(
            repo,
            name,
            host=host,
            version=version,
            pre_hook=pre_hook,
            post_hook=post_hook,
            batch_size=batch_size,
            log_to_field=log_to_field,
        )
        self.datasource.metadata_field(log_to_field).set_annotation().apply()
        return res

    def predict_with_mlflow_model(
        self,
        repo: str,
        name: str,
        host: Optional[str] = None,
        version: str = "latest",
        pre_hook: Callable[[List[str]], Any] = identity_func,
        post_hook: Callable[[Any], Any] = identity_func,
        batch_size: int = 1,
        log_to_field: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Fetch an MLflow model from a specific repository and use it to predict on the datapoints in this QueryResult.

        Any MLflow model that has a ``model.predict`` endpoint is supported.
        This includes, but is not limited to the following flavors:

        * ``torch``
        * ``tensorflow``
        * ``pyfunc``
        * ``scikit-learn``

        Keep in mind that by default ``mlflow.predict()`` will receive the list of downloaded datapoint paths as input,
        so additional "massaging" of the data might be required for prediction to work.
        Use the ``pre_hook`` function to do so.

        Args:
            repo: repository to extract the model from

            name: name of the model in the repository's MLflow registry.
            host: address of the DagsHub instance with the repo to load the model from.
                 Set it if the model is hosted on a different DagsHub instance than the datasource.
            version: version of the model in the mlflow registry.
            pre_hook: function that runs before datapoints are sent to ``model.predict()``.
                The input argument is the list of paths to datapoint files in the current batch.
            post_hook: function that converts the model output to the desired format.
            batch_size: Size of the file batches that are sent to ``model.predict()``.
                Default batch size is 1, but it is still being sent as a list for consistency.
            log_to_field: If set, writes prediction results to this metadata field in the datasource.
        """
        if not host:
            host = self.datasource.source.repoApi.host

        prev_uri = mlflow.get_tracking_uri()
        mlflow.set_tracking_uri(multi_urljoin(host, f"{repo}.mlflow"))
        token = get_token(host=host)
        os.environ["MLFLOW_TRACKING_USERNAME"] = UserAPI.get_user_from_token(token, host=host).username
        os.environ["MLFLOW_TRACKING_PASSWORD"] = token
        model_uri = f"models:/{name}/{version}"

        try:
            loader_module = mlflow.models.get_model_info(model_uri).flavors["python_function"]["loader_module"]
            loader_module_elems = loader_module.split(".")
            if loader_module_elems[-1] == "model":
                loader_module_elems.pop()
            loader_module = ".".join(loader_module_elems)
            loader = mlflow.pyfunc if "pyfunc" in loader_module_elems else importlib.import_module(loader_module)
            model = loader.load_model(model_uri)
        finally:
            mlflow.set_tracking_uri(prev_uri)

        if "torch" in loader_module:
            model.predict = model.__call__

        return self.generate_predictions(lambda x: post_hook(model.predict(pre_hook(x))), batch_size, log_to_field)

    def get_annotations(self, **kwargs) -> "QueryResult":
        """
        Loads all annotation fields using :func:`get_blob_fields`.

        All keyword arguments are passed to :func:`get_blob_fields`.
        """
        if len(self.annotation_fields) == 0:
            return self

        log_message(f"Downloading annotation fields {self.annotation_fields}...")
        self.get_blob_fields(*self.annotation_fields, **kwargs)

        return self

    def download_files(
        self,
        target_dir: Optional[Union[str, PathLike]] = None,
        keep_source_prefix=True,
        redownload=False,
        path_field: Optional[str] = None,
    ) -> PathLike:
        """
        Downloads the datapoints to the ``target_dir`` directory

        Args:
            target_dir: Where to download the files. Defaults to \
                :func:`datasource's default location \
                <dagshub.data_engine.model.datasource.Datasource.default_dataset_location>`
            keep_source_prefix: If True, includes the prefix of the datasource in the download path.
            redownload: Whether to redownload a file if it exists on the filesystem already.\
                We don't do any hashsum checks, so if it's possible that the file has been updated, set to True
            path_field: Set this to the name of the field with the file's path\
                if you want to download files from a field other than the datapoint's path.

        .. note::
            For ``path_field`` the path in the field still needs to be in the same repo
            and have the same format as the path of the datapoint, including not having the prefix.
            For now, you can't download arbitrary paths/urls.

        Returns:
            Path to the directory with the downloaded files
        """

        target_path = self.datasource.default_dataset_location if target_dir is None else Path(target_dir).expanduser()
        logger.warning(f"Downloading {len(self.entries)} files to {str(target_path)}")

        if self.datasource.source.source_type == DatasourceType.BUCKET:
            send_analytics_event(
                "Client_DataEngine_downloadDatapointsFromBucket",
                repo=self.datasource.source.repoApi,
            )
        else:
            send_analytics_event(
                "Client_DataEngine_downloadDatapoints",
                repo=self.datasource.source.repoApi,
            )

        def dp_path(dp: Datapoint):
            if path_field is not None:
                path_val = dp.metadata.get(path_field)
                if path_val is None:
                    return None
            else:
                path_val = dp.path

            if keep_source_prefix:
                return target_path / self.datasource.source.source_prefix / path_val
            else:
                return target_path / path_val

        def dp_url(dp: Datapoint):
            if path_field is not None:
                path_val = dp.metadata.get(path_field)
                if path_val is None:
                    return None
                return self.datasource.source.raw_path(path_val)
            else:
                return dp.download_url

        download_args = [(dp_url(dp), dp_path(dp)) for dp in self.entries if dp_path(dp) is not None]

        download_files(download_args, skip_if_exists=not redownload)
        return target_path

    def _get_all_annotations(self, annotation_field: str) -> List[IRImageAnnotationBase]:
        annotations = []
        for dp in self.entries:
            if annotation_field in dp.metadata:
                if not hasattr(dp.metadata[annotation_field], "annotations"):
                    print(f"Skipping datapoint {dp.path} since it doesn't have correct annotations")
                    continue
                annotations.extend(dp.metadata[annotation_field].annotations)
        return annotations

    def export_as_yolo(
        self,
        download_dir: Optional[Union[str, Path]] = None,
        annotation_field: Optional[str] = None,
        annotation_type: Optional[Literal["bbox", "segmentation", "pose"]] = None,
        classes: Optional[Dict[int, str]] = None,
    ) -> Path:
        """
        Downloads the files and annotations in a way that can be used to train with YOLO immediately.

        Args:
            download_dir: Where to download the files. Defaults to ``./dagshub_export``
            annotation_field: Field with the annotations. If None, uses the first alphabetical annotation field.
            annotation_type: Type of YOLO annotations to export.
                Possible values: "bbox", "segmentation", "pose".
                If None, returns based on the most common annotation type.
            classes: Classes and indices for the YOLO dataset.
                If ``None``, the classes will be inferred from the annotations.
                Any class in the annotations that doesn't exist in the dictionary
                will be added at the end of the classes.
                The dictionary has to be in the format ``{index: class_name}``.

        Returns:
            The path to the YAML file with the metadata. Pass this path to ``YOLO.train()`` to train a model.
        """
        if annotation_field is None:
            annotation_fields = sorted([f.name for f in self.fields if f.is_annotation()])
            if len(annotation_fields) == 0:
                raise ValueError("No annotation fields found in the datasource")
            annotation_field = annotation_fields[0]
            log_message(f"Using annotations from field {annotation_field}")

        if download_dir is None:
            download_dir = Path("dagshub_export")
        download_dir = Path(download_dir) / "data"

        annotations = self._get_all_annotations(annotation_field)

        categories = Categories()
        if classes is not None:
            for idx, cls in classes.items():
                categories.add(cls, idx)
        ann_types: dict[type, int] = {}

        for ann in annotations:
            if annotation_type is None:
                ann_type = type(ann)
                ann_types[ann_type] = ann_types.get(ann_type, 0) + 1
            for cat in ann.categories.keys():
                categories.get_or_create(cat)

        if annotation_type is None:
            annotation_types = [type(ann) for ann in annotations]
            counter = Counter(annotation_types)
            most_common = counter.most_common(1)
            if not most_common:
                raise RuntimeError("No annotations found")
            most_common_type = most_common[0][0]
            annotation_type = ir_mapping.get(most_common_type)
            if annotation_type is None:
                raise RuntimeError(f"Don't know how to export annotations of type {most_common_type} to YOLO")
            log_message(f"Importing annotations of type {annotation_type} (most common type of annotation).")

        image_download_path = Path(download_dir)

        # Add the source prefix to all annotations
        for ann in annotations:
            ann.filename = os.path.join(self.datasource.source.source_prefix, ann.filename)

        # If there's no folder "images" in the datasource, prepend it to the path
        if not any("images/" in ann.filename for ann in annotations):
            image_download_path = Path(download_dir) / "images"

        context = YoloContext(annotation_type=annotation_type, categories=categories)
        context.path = image_download_path

        log_message("Downloading image files...")
        self.download_files(image_download_path)
        log_message("Exporting annotations...")
        yaml_path = dagshub_annotation_converter.converters.yolo.export_to_fs(context, annotations)
        log_message(f"Done! Saved YOLO Dataset, YAML file is at {yaml_path.absolute()}")
        return yaml_path

    def to_voxel51_dataset(self, **kwargs) -> "fo.Dataset":
        """
        Creates a voxel51 dataset that can be used with\
        `fo.launch_app()
        <https://docs.voxel51.com/api/fiftyone.core.session.html?highlight=launch_app#fiftyone.core.session.launch_app>`_
        to visualize it.

        Keyword Args:
            name (str): Name of the dataset. Default is the name of the datasource.
            force_download (bool): Download the dataset even if the size of the files is bigger than 100MB.\
                Default is False
            files_location (str|PathLike): path to the location where to download the local files.
                Default is :func:`datasource's default location \
                <dagshub.data_engine.model.datasource.Datasource.default_dataset_location>`
            redownload (bool): Redownload files, replacing the ones that might exist on the filesystem.\
                Default is False.
            voxel_annotations (List[str]) : List of fields from which to load voxel annotations serialized with
                                        `to_json()`. This will override the labelstudio annotations

        :rtype: `fo.Dataset <https://docs.voxel51.com/api/fiftyone.core.dataset.html#fiftyone.core.dataset.Dataset>`_
        """
        if len(self.entries) == 0:
            raise VisualizeError("No datapoints to visualize")
        logger.info("Migrating dataset to voxel51")
        name = kwargs.get("name", self.datasource.source.name)
        force_download = kwargs.get("force_download", False)
        # TODO: don't override samples in existing one
        if fo.dataset_exists(name):
            ds: fo.Dataset = fo.load_dataset(name)
        else:
            ds: fo.Dataset = fo.Dataset(name)
        # ds.persistent = True

        dataset_location = Path(kwargs.get("files_location", self.datasource.default_dataset_location))
        dataset_location.mkdir(parents=True, exist_ok=True)
        logger.info("Downloading files...")

        if "voxel_annotations" in kwargs:
            annotation_fields = kwargs["voxel_annotations"]
            label_func = add_voxel_annotations
        else:
            annotation_fields = [f for f in self.fields if f.is_annotation()]
            label_func = add_ls_annotations

        # Load the annotation fields
        self.download_binary_columns(*annotation_fields)

        if not force_download:
            self._check_downloaded_dataset_size()

        redownload = kwargs.get("redownload", False)
        self.download_files(self.datasource.default_dataset_location, redownload=redownload)

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())
        task = progress.add_task("Generating voxel samples...", total=len(self.entries))

        samples: List["fo.Sample"] = []

        with progress:
            for datapoint in self.entries:
                filepath = self.datasource.default_dataset_location / datapoint.path_in_repo
                sample = fo.Sample(filepath=filepath)
                sample["dagshub_download_url"] = datapoint.download_url
                sample["datapoint_id"] = datapoint.datapoint_id
                sample["datapoint_path"] = datapoint.path
                label_func(sample, datapoint, *annotation_fields)
                for k, v in datapoint.metadata.items():
                    # TODO: more filtering here, not all fields should be showing up in voxel
                    if k in annotation_fields:
                        continue
                    if type(v) is not bytes:
                        sample[k] = v
                samples.append(sample)
                progress.update(task, advance=1, refresh=True)

        ds.merge_samples(samples)
        return ds

    @staticmethod
    def _get_predict_dict(predictions, remote_path, log_to_field):
        res = {log_to_field: json.dumps(predictions[remote_path][0]).encode("utf-8")}
        if len(predictions[remote_path]) == 2:
            res[f"{log_to_field}_score"] = predictions[remote_path][1]

        return res

    def _check_downloaded_dataset_size(self):
        download_size_prompt_threshold = 100 * (2**20)  # 100 Megabytes
        dp_size = self._calculate_datapoint_size()
        if dp_size is not None and dp_size > download_size_prompt_threshold:
            prompt = f"You're about to download {sizeof_fmt(dp_size)} of images locally."
            should_download = prompt_user(prompt)
            if not should_download:
                msg = "Downloading voxel dataset cancelled"
                logger.warning(msg)
                raise RuntimeError(msg)

    def _calculate_datapoint_size(self) -> Optional[int]:
        sum_size = 0
        has_sum_field = False
        all_have_sum_field = True
        size_field = "size"
        for dp in self.entries:
            if size_field in dp.metadata:
                has_sum_field = True
                sum_size += dp.metadata[size_field]
            else:
                all_have_sum_field = False
        if not has_sum_field:
            logger.warning("None of the datapoints had a size field, can't calculate size of the downloading dataset")
            return None
        if not all_have_sum_field:
            logger.warning("Not every datapoint has a size field, size calculations might be wrong")
        return sum_size

    def visualize(self, visualizer: Literal["dagshub", "fiftyone"] = "dagshub", **kwargs) -> Union[str, "fo.Session"]:
        """
        Visualize this QueryResult either on DagsHub or with Voxel51.

        If ``visualizer`` is ``dagshub``, a webpage is opened on DagsHub with the query applied.

        If ``visualizer`` is ``fiftyone``, this function calls :func:`to_voxel51_dataset`,
        passing to it the keyword arguments, and launches a fiftyone session showing the dataset.

        Additionally, this function adds a DagsHub plugin into Voxel51 that you can use for additional interactions
        with the datasource from within the voxel environment.

        Returns the session object that you can ``wait()`` if you are using it\
        outside a notebook and need to not close the script immediately::

            session = ds.fetch().visualize()
            session.wait(-1)
        """
        if visualizer == "dagshub":
            return self.datasource.visualize(visualizer, **kwargs)
        if visualizer == "fiftyone":
            set_voxel_envvars()

            send_analytics_event("Client_DataEngine_VizualizeResults", repo=self.datasource.source.repoApi)

            ds = self.to_voxel51_dataset(**kwargs)

            sess = fo.launch_app(ds)
            # Launch the server for plugin interaction
            plugin_server_module.run_plugin_server(sess, self.datasource, self.datasource.source.revision)

            return sess

    def generate_predictions(
        self,
        predict_fn: CustomPredictor,
        batch_size: int = 1,
        log_to_field: Optional[str] = None,
    ) -> Dict[str, Tuple[str, Optional[float]]]:
        """
        Sends all the datapoints returned in this QueryResult as prediction targets for
        a generic object.

        Args:
            predict_fn: function that handles batched input and returns predictions with an optional prediction score.
            batch_size: (optional, default: 1) number of datapoints to run inference on simultaneously
            log_to_field: (optional, default: 'prediction') write prediction results to metadata logged in data engine.
            If None, just returns predictions.
            (in addition to logging to a field, iff that parameter is set)
        """
        dset = DagsHubDataset(self, tensorizers=[lambda x: x])

        predictions = {}
        progress = get_rich_progress(rich.progress.MofNCompleteColumn())
        task = progress.add_task("Running inference...", total=len(dset))
        with progress:
            for idx, local_paths in enumerate(
                _Batcher(dset, batch_size) if batch_size != 1 else dset
            ):  # encapsulates dataset with batcher if necessary and iterates over it
                for prediction, remote_path in zip(
                    predict_fn(local_paths),
                    [result.path for result in self[idx * batch_size : (idx + 1) * batch_size]],
                ):
                    predictions[remote_path] = prediction
                progress.update(task, advance=batch_size, refresh=True)

        if log_to_field:
            with self.datasource.metadata_context() as ctx:
                for remote_path in predictions:
                    ctx.update_metadata(remote_path, self._get_predict_dict(predictions, remote_path, log_to_field))
        return predictions

    def generate_annotations(self, predict_fn: CustomPredictor, batch_size: int = 1, log_to_field: str = "annotation"):
        """
        Sends all the datapoints returned in this QueryResult as prediction targets for
        a generic object.

        Args:
            predict_fn: function that handles batched input and returns predictions with an optional prediction score.
            batch_size: (optional, default: 1) number of datapoints to run inference on simultaneously.
            log_to_field: (optional, default: 'prediction') write prediction results to metadata logged in data engine.
        """
        self.generate_predictions(
            predict_fn,
            batch_size=batch_size,
            log_to_field=log_to_field,
        )
        self.datasource.metadata_field(log_to_field).set_annotation().apply()

    def annotate(
        self,
        open_project: bool = True,
        ignore_warning: bool = True,
        fields_to_embed: Optional[List[str]] = None,
        fields_to_exclude: Optional[List[str]] = None,
    ) -> Optional[str]:
        """
        Sends all the datapoints returned in this QueryResult to be annotated in Label Studio on DagsHub.
        Alternatively, uses MLFlow to automatically label datapoints.

        Args:
            open_project: Automatically open the Label Studio project in the browser
            ignore_warning: Suppress the prompt-warning if you try to annotate too many datapoints at once.
            fields_to_embed: list of meta-data columns that will show up in Label Studio UI.
             if not specified all will be displayed.
            fields_to_exclude: list of meta-data columns that will not show up in Label Studio UI
        Returns:
            The URL of the created Label Studio workspace
        """
        send_analytics_event("Client_DataEngine_SentToAnnotation", repo=self.datasource.source.repoApi)

        return self.datasource.send_datapoints_to_annotation(
            self.entries,
            open_project=open_project,
            ignore_warning=ignore_warning,
            fields_to_exclude=fields_to_exclude,
            fields_to_embed=fields_to_embed,
        )

    @property
    def document_fields(self) -> List[str]:
        return [f.name for f in self.fields if f.is_document()]

    @property
    def annotation_fields(self) -> List[str]:
        return [f.name for f in self.fields if f.is_annotation()]

    def _load_autoload_fields(self, documents=True, annotations=True):
        """
        Loads fields that are supposed to be load automatically upon querying.
        This includes:
            - All document fields
            - All annotation fields
        """
        if documents:
            if len(self.document_fields) > 0:
                self.get_blob_fields(*self.document_fields, load_into_memory=True)

        if annotations:
            self.get_annotations()

    def log_to_mlflow(self, run: Optional["mlflow.entities.Run"] = None) -> "mlflow.entities.Run":
        """
        Logs the query result information to MLflow as an artifact.
        The artifact will be saved at the root of the run with the name in the format of
        ``log_{datasource_name}_{query_time}_{random_chunk}.dagshub.dataset.json``.

        You can later load the dataset back from MLflow using :func:`dagshub.data_engine.datasources.get_from_mlflow`.

        Args:
            run: MLflow run to save to. If ``None``, uses the active MLflow run or creates a new run.

        Returns:
            Run to which the artifact was logged.
        """
        assert self.query_data_time is not None
        artifact_name = self.datasource._get_mlflow_artifact_name("log", self.query_data_time)
        return self.datasource._log_to_mlflow(artifact_name, run, self.query_data_time)


# to support depedency-free dataloading, `_Batcher` is a barebones dataloader that sets up batched inference
class _Batcher:
    def __init__(self, dset, batch_size):
        self.dset = dset
        self.batch_size = batch_size

    def __iter__(self):
        self.curr_idx = 0
        return self

    def __next__(self):
        self.curr_idx += self.batch_size
        return [self.dset[idx] for idx in range(self.curr_idx - self.batch_size, self.curr_idx)]



from tenacity import retry, wait_fixed, stop_after_attempt, retry_if_exception_type
from json import JSONDecodeError
from typing import Optional
import importlib.util
import semver

from dagshub.common.api.repo import RepoAPI
from dagshub.auth import get_token
from dagshub.common import config


def _use_legacy_client():
    """
    https://github.com/HumanSignal/label-studio/releases/tag/1.13.0, \
            https://github.com/HumanSignal/label-studio/pull/5961; \
            introduces breaking changes; anyone using SDK < 1.0 should use the legacy client.
    :meta experimental:
    """
    import label_studio_sdk

    return semver.compare("1.0.0", label_studio_sdk.__version__) == 1


@retry(retry=retry_if_exception_type(JSONDecodeError), wait=wait_fixed(3), stop=stop_after_attempt(5))
def get_label_studio_client(
    repo: str, legacy_client: Optional[bool] = None, host: Optional[str] = None, token: Optional[str] = None
):
    """
    Creates a
    `label_studio_sdk.client.LabelStudio <https://api.labelstud.io/api-reference/introduction/getting-started>`_ /
    `label_studio_sdk.Client (legacy) <https://labelstud.io/guide/sdk>`_
    object to interact with the LabelStudio instance associated with the repository.

    Args:
        repo: Name of the repo in the format of ``username/reponame``
        legacy_client: if True, returns the older legacy LabelStudio Client.
        host: URL of the hosted DagsHub instance. default is ``https://dagshub.com``.
        token: (optional, default: None) use this token for LS requests. By default, will use current user's token.

    Returns:
        `label_studio_sdk.Client` / `label_studio_sdk.client.LabelStudio` object
    """

    if importlib.util.find_spec("label_studio_sdk") is None:
        raise ModuleNotFoundError("Could not import module label_studio_sdk. Make sure to pip install label_studio_sdk")

    if legacy_client is None:
        legacy_client = _use_legacy_client()

    if not host:
        host = config.host

    if legacy_client:
        from label_studio_sdk import Client as LabelStudio
    else:
        from label_studio_sdk.client import LabelStudio

    repo_api = RepoAPI(repo, host=host)
    kwargs = {
        "url" if legacy_client else "base_url": repo_api.label_studio_api_url()[:-4],
        "api_key": token if token is not None else get_token(host=host),
    }

    return LabelStudio(**kwargs)



import datetime
import logging
import math
from typing import List, TYPE_CHECKING, Dict, Any, Optional

from dagshub.common.helpers import log_message
from dagshub.common.util import lazy_load, wrap_bytes
from dagshub.data_engine import dtypes
from dagshub.data_engine.client.models import autogenerated_columns
from dagshub.data_engine.dtypes import MetadataFieldType
from dagshub.data_engine.model.metadata.dtypes import DatasourceFieldInfo, DatapointMetadataUpdateEntry
from dagshub.data_engine.model.metadata.util import _get_datetime_utc_offset
from dagshub.data_engine.model.metadata.validation import UploadingMetadataInfo, MAX_STRING_FIELD_LENGTH
from dagshub.data_engine.model.schema_util import metadata_type_lookup, special_metadata_handlers

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource
    import pandas
else:
    pandas = lazy_load("pandas")

logger = logging.getLogger(__name__)


def _transform_strings_to_documents(ds: "Datasource", metadata: List[DatapointMetadataUpdateEntry], field_name: str):
    log_message(f"Uploading field {field_name} as a document field due to long string values")
    ds.metadata_field(field_name).set_type(dtypes.Document).apply()
    for m in metadata:
        if m.key != field_name:
            continue
        m.valueType = MetadataFieldType.BLOB
        m.value = wrap_bytes(m.value.encode("utf-8"))


def run_preupload_transforms(
    ds: "Datasource",
    metadata: List[DatapointMetadataUpdateEntry],
    precalculated_info: Dict[str, UploadingMetadataInfo],
):
    """
    Run transformations on metadata that should happen before uploading.

    Current transformations:
        - New string fields with large values get converted to a Document type
    """

    for info in precalculated_info.values():
        if (
            info.field_type == MetadataFieldType.STRING
            and info.existing_metadata_in_ds is None
            and len(info.longest_value) > MAX_STRING_FIELD_LENGTH
        ):
            _transform_strings_to_documents(ds, metadata, info.field_name)


def _add_metadata(
    field_info: DatasourceFieldInfo,
    buf: List[DatapointMetadataUpdateEntry],
    dp_path: str,
    key: str,
    value: Any,
    is_pandas: bool,
):
    """
    Adds the new metadata to the buf buffer.
    Have to pass in the buffer in order to graduate the already added metadata to multivalue, if that happens.
    """
    if key in autogenerated_columns:
        return
    if value is None:
        return
    if is_pandas:
        # Pandas specific: since pandas doesn't distinguish between None and NaN, don't upload it
        if isinstance(value, float) and math.isnan(value):
            return
        # Pandas specific: Cast pandas Timestamp to datetime
        if isinstance(value, list) or isinstance(value, pandas._libs.tslibs.timestamps.Timestamp):
            if isinstance(value, list):
                if isinstance(value[0], pandas._libs.tslibs.timestamps.Timestamp):
                    value = [val.to_pydatetime() for val in value]
            else:
                value = value.to_pydatetime()

    def _generate_metadata_update_entry(val: Any, is_multivalue: bool) -> Optional[DatapointMetadataUpdateEntry]:
        field_type = field_info.field_value_types.get(key)
        value_type = type(val)

        # Run special handling first
        if value_type in special_metadata_handlers:
            val, converted_field_type = special_metadata_handlers[value_type](val)
            if val is None:
                return None
            if field_type is not None and converted_field_type != field_type:
                raise ValueError(f"Can't put value of type {value_type} into field of type {field_type}")
            return DatapointMetadataUpdateEntry(
                url=dp_path,
                key=key,
                value=val,
                valueType=converted_field_type,
                allowMultiple=is_multivalue,
                timeZone=None,
            )

        tz = None
        if field_type is None:
            field_type = metadata_type_lookup[type(val)]
            field_info.field_value_types[key] = field_type
        # Don't override bytes if they're not bytes - probably just undownloaded values
        if field_type == MetadataFieldType.BLOB and not isinstance(val, bytes):
            if key not in field_info.document_fields:
                return None
        # Pandas quirk - integers are floats on the backend
        if is_pandas:
            if field_type == MetadataFieldType.INTEGER:
                val = int(val)
        if isinstance(val, str) and key in field_info.document_fields:
            val = val.encode("utf-8")
        if isinstance(val, bytes):
            val = wrap_bytes(val)
        if isinstance(val, datetime.datetime):
            tz = _get_datetime_utc_offset(val)
            val = int(val.timestamp() * 1000)
        return DatapointMetadataUpdateEntry(
            url=dp_path,
            key=key,
            value=str(val),
            valueType=field_type,
            allowMultiple=is_multivalue,
            timeZone=tz,
        )

    if isinstance(value, list):
        if key not in field_info.multivalue_fields:
            field_info.multivalue_fields.add(key)
            # Promote all the existing uploading metadata to multivalue
            for update_entry in buf:
                if update_entry.key == key:
                    update_entry.allowMultiple = True
        for sub_val in value:
            metadata_entry = _generate_metadata_update_entry(sub_val, True)
            if metadata_entry is not None:
                buf.append(metadata_entry)
    else:
        metadata_entry = _generate_metadata_update_entry(value, key in field_info.multivalue_fields)
        if metadata_entry is not None:
            buf.append(metadata_entry)



from typing import TYPE_CHECKING, Optional, Sequence, Tuple, Union, Literal, Dict

from dagshub_annotation_converter.formats.label_studio.task import parse_ls_task, LabelStudioTask
from dagshub_annotation_converter.formats.yolo import import_lookup, import_yolo_result, YoloContext
from dagshub_annotation_converter.formats.yolo.categories import Categories
from dagshub_annotation_converter.ir.image import (
    IRBBoxImageAnnotation,
    CoordinateStyle,
    IRSegmentationImageAnnotation,
    IRSegmentationPoint,
    IRPoseImageAnnotation,
    IRPosePoint,
)
from dagshub_annotation_converter.ir.image.annotations.base import IRAnnotationBase, IRImageAnnotationBase

from dagshub.common.api import UserAPI
from dagshub.common.helpers import log_message

if TYPE_CHECKING:
    from dagshub.data_engine.model.datapoint import Datapoint
    import ultralytics.engine.results


class AnnotationMetaDict(dict):
    def __init__(self, annotation: "MetadataAnnotations", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.annotation = annotation

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.annotation._update_datapoint()


class MetadataAnnotations:
    """
    Class that holds metadata annotations for a datapoint.

    This class is automatically created for every datapoint,
    as long as the field is a blob field and
    :func:`has been marked as annotation \
    <dagshub.data_engine.model.metadata_field_builder.MetadataFieldBuilder.set_annotation>`


    Example of adding bounding boxes::

        dp = ds.fetch()[0]
        anns: MetadataAnnotations = dp["exported_annotations"]
        anns.add_image_bbox("person", 0.1, 0.1, 0.1, 0.1)
        anns.add_image_bbox("cat", 0.2, 0.2, 0.1, 0.1)
        anns.meta["some_key"] = "some_value"
        dp.save()

    You can use the ``meta`` dictionary to add additional metadata to the task, as long as it is JSON-serializable.

    All functions for adding annotations have additional arguments of ``image_width``/``image_height``.
    They are required for new datapoints, but if there are already annotations existing,
    or if there is width/height in the metadata, they can be omitted.
    """

    def __init__(
        self,
        datapoint: "Datapoint",
        field: str,
        annotations: Optional[Sequence["IRAnnotationBase"]] = None,
        meta: Optional[Dict] = None,
        original_value: Optional[bytes] = None,
    ):
        self.datapoint = datapoint
        self.field = field
        self.annotations: list["IRAnnotationBase"]
        if annotations is None:
            annotations = []
        self.annotations = list(annotations)
        self.meta = AnnotationMetaDict(self, meta or {})
        self._original_value = original_value

    @staticmethod
    def from_ls_task(datapoint: "Datapoint", field: str, ls_task: bytes) -> "MetadataAnnotations":
        parsed_ls_task = parse_ls_task(ls_task)
        annotations = parsed_ls_task.to_ir_annotations(filename=datapoint.path)

        return MetadataAnnotations(
            datapoint=datapoint, field=field, annotations=annotations, meta=parsed_ls_task.meta, original_value=ls_task
        )

    def to_ls_task(self) -> Optional[bytes]:
        """
        Convert the annotations into a Label Studio task (this is what's stored in the Data Engine backend).
        """
        if len(self.annotations) == 0:
            return None
        task = LabelStudioTask(
            user_id=UserAPI.get_current_user(self.datapoint.datasource.source.repoApi.host).user_id,
        )
        task.data["image"] = self.datapoint.download_url
        # TODO: need to filter out non-image annotations here maybe?
        task.add_ir_annotations(self.annotations)
        task.meta.update(self.meta)
        return task.model_dump_json().encode("utf-8")

    @property
    def value(self) -> Optional[bytes]:
        """
        Returns the contents of annotation as a byte array.
        If it was loaded from the backend and not changed, it will return the original value.

        If there were any changes, it will instead return the serialized version of the annotations
        (the username will be set to the current user).
        """
        if self._original_value is not None:
            return self._original_value
        return self.to_ls_task()

    def __repr__(self):
        return f"Annotations:\n\t{self.annotations}"

    def get_image_dimensions(self, image_width: Optional[int], image_height: Optional[int]) -> Tuple[int, int]:
        """
        Get dimensions of the image.
        If not provided, tries to get them from the existing annotations or the datapoint metadata.

        :meta private:
        """
        if image_width is not None and image_height is not None:
            return image_width, image_height

        for ann in self.annotations:
            if isinstance(ann, IRImageAnnotationBase):
                return ann.image_width, ann.image_height

        if "width" not in self.datapoint.metadata:
            raise ValueError('Image width not provided, and a "width" field was not found in the datapoint')
        if "height" not in self.datapoint.metadata:
            raise ValueError('Image height not provided, and a "height" field was not found in the datapoint')
        return self.datapoint["width"], self.datapoint["height"]

    def _update_datapoint(self):
        """
        Fire this method on every update to save annotations in the datapoint
        """
        self.datapoint[self.field] = self
        self._original_value = self.to_ls_task()

    def add_image_bbox(
        self,
        category: str,
        top: float,
        left: float,
        width: float,
        height: float,
        image_width: Optional[int] = None,
        image_height: Optional[int] = None,
    ):
        """
        Adds a bounding box annotation.
        Values need to be normalized from 0 to 1

        Args:
            category: Annotation category
            top: Top coordinate of the bounding box
            left: Left coordinate of the bounding box
            width: Width of the bounding box
            height: Height of the bounding box
            image_width: Width of the image. If not supplied, tries to get it from the `width` field in datapoint
            image_height: Height of the image. If not supplied, tries to get it from the `height` field in datapoint
        """
        image_width, image_height = self.get_image_dimensions(image_width, image_height)
        self.annotations.append(
            IRBBoxImageAnnotation(
                filename=self.datapoint.path,
                categories={category: 1.0},
                top=top,
                left=left,
                width=width,
                height=height,
                image_width=image_width,
                image_height=image_height,
                coordinate_style=CoordinateStyle.NORMALIZED,
            )
        )
        self._update_datapoint()

    def add_image_segmentation(
        self,
        category: str,
        points: Sequence[Tuple[float, float]],
        image_width: Optional[int] = None,
        image_height: Optional[int] = None,
    ):
        """
        Add a segmentation annotation.
        Points need to be a list of tuples of 2 (x, y) values, normalized from 0 to 1.
        Example of points: ``[(0.1, 0.1), (0.3, 0.3), (0.1, 0.6)]``


        Args:
            category: Annotation category
            points: List of points of the segmentation
            image_width: Width of the image. If not supplied, tries to get it from the `width` field in datapoint
            image_height: Height of the image. If not supplied, tries to get it from the `height` field in datapoint
        """
        image_width, image_height = self.get_image_dimensions(image_width, image_height)
        self.annotations.append(
            IRSegmentationImageAnnotation(
                filename=self.datapoint.path,
                categories={category: 1.0},
                points=[IRSegmentationPoint(x=x, y=y) for x, y in points],
                image_width=image_width,
                image_height=image_height,
                coordinate_style=CoordinateStyle.NORMALIZED,
            )
        )
        self._update_datapoint()

    def add_image_pose(
        self,
        category: str,
        points: Union[Sequence[Tuple[float, float]], Sequence[Tuple[float, float, Optional[bool]]]],
        bbox_left: Optional[float] = None,
        bbox_top: Optional[float] = None,
        bbox_width: Optional[float] = None,
        bbox_height: Optional[float] = None,
        image_width: Optional[int] = None,
        image_height: Optional[int] = None,
    ):
        """
        Adds a new pose annotation

        ``bbox_...`` arguments define the bounding box of the pose. If any of the parameters is not defined,
        the bounding box is instead created from the points.

        Points need to be a list of tuples of ``(x, y)`` or ``(x, y, visible)`` values, normalized from 0 to 1.

        Args:
            category: Annotation category
            points: List of points of the pose
            bbox_left: Left coordinate of the bounding box
            bbox_top: Top coordinate of the bounding box
            bbox_width: Width of the bounding box
            bbox_height: Height of the bounding box
            image_width: Width of the image. If not supplied, tries to get it from the `width` field in datapoint
            image_height: Height of the image. If not supplied, tries to get it from the `height` field in datapoint
        """

        image_width, image_height = self.get_image_dimensions(image_width, image_height)

        pose_points: list[IRPosePoint] = []
        for p in points:
            if len(p) == 2:
                pose_points.append(IRPosePoint(x=p[0], y=p[1]))
            else:
                pose_points.append(IRPosePoint(x=p[0], y=p[1], visible=p[2]))

        ann = IRPoseImageAnnotation.from_points(
            filename=self.datapoint.path,
            categories={category: 1.0},
            points=pose_points,
            image_width=image_width,
            image_height=image_height,
            coordinate_style=CoordinateStyle.NORMALIZED,
        )

        if bbox_left is not None and bbox_top is not None and bbox_width is not None and bbox_height is not None:
            ann.left = bbox_left
            ann.top = bbox_top
            ann.width = bbox_width
            ann.height = bbox_height

        self.annotations.append(ann)
        self._update_datapoint()

    def add_yolo_annotation(
        self,
        annotation_type: Literal["bbox", "segmentation", "pose"],
        annotation: Union[str, "ultralytics.engine.results.Results"],
        categories: Optional[Dict[int, str]] = None,
        image_width: Optional[int] = None,
        image_height: Optional[int] = None,
        pose_keypoint_dim: Optional[int] = None,
    ):
        """
        Add a YOLO annotation from string or from a result of prediction with a YOLO model.

        This could be either a string of an annotations from a YOLO file, or a result of evaluating a YOLO model.
        Args:

        """
        annotations: list[IRAnnotationBase] = []
        if isinstance(annotation, str):
            if categories is None:
                raise ValueError("`categories` dictionary is required when importing annotations from a string")
            image_width, image_height = self.get_image_dimensions(image_width, image_height)
            yolo_context = self._generate_yolo_context(annotation_type, categories)
            if annotation_type == "pose":
                if pose_keypoint_dim is None:
                    raise ValueError("`pose_keypoint_dim` is required when importing pose annotations")
                yolo_context.keypoint_dim = pose_keypoint_dim
            parse_fn = import_lookup[annotation_type]
            for ann in annotation.split("\n"):
                new_ann = parse_fn(ann, yolo_context, image_width, image_height, None)
                new_ann.filename = self.datapoint.path
                annotations.append(new_ann)
        else:
            new_anns = import_yolo_result(annotation_type, annotation)
            for new_ann in new_anns:
                new_ann.filename = self.datapoint.path
            annotations.extend(new_anns)
        self.annotations.extend(annotations)
        log_message(f"Added {len(annotations)} YOLO annotation(s) to datapoint {self.datapoint.path}")
        self._update_datapoint()

    @staticmethod
    def _generate_yolo_context(annotation_type, categories: Dict[int, str]) -> YoloContext:
        cats = Categories()
        for cat_id, cat_name in categories.items():
            cats.add(cat_name, cat_id)
        return YoloContext(annotation_type=annotation_type, categories=cats)



import datetime
import json
import logging
import tempfile
from pathlib import Path, PurePosixPath
from socket import gethostname, gethostbyname
from typing import TYPE_CHECKING

import httpx
import urllib.parse

from dagshub.common.util import lazy_load
from dagshub.common.helpers import log_message
from dagshub.upload import Repo

if TYPE_CHECKING:
    import IPython as IPy
else:
    IPy = lazy_load("IPython")


logger = logging.getLogger(__name__)


def _inside_notebook():
    return IPy.get_ipython() is not None


def _inside_colab():
    try:
        if IPy.get_ipython() and "google.colab" in IPy.get_ipython().extension_manager.loaded:
            return True
    except Exception:
        pass
    return False


def _default_notebook_name():
    if _inside_colab():
        # Shout out to https://stackoverflow.com/a/61906730
        ip = gethostbyname(gethostname())
        filename = httpx.get(f"http://{ip}:9000/api/sessions").json()[0]["name"]
        filename = urllib.parse.unquote(filename)
        if not filename.endswith(".ipynb"):
            filename = filename + ".ipynb"
        return filename
    return f"notebook-{datetime.datetime.utcnow().strftime('%Y-%m-%d')}.ipynb"


def save_notebook(repo, path="", branch=None, commit_message=None, versioning="git", colab_timeout=40) -> None:
    """
    Save the notebook to DagsHub.

    Args:
        repo: Repository in the format of ``user/repo``.
        path: Path of the notebook in repo, including the filename.
             If left empty, saves the notebook to the root of the repo with format ``notebook-{date.now}.ipynb``.
             If path is a directory and not a file (no extension), saves it to ``path/notebook-{date.now}.ipynb``.
        branch: The branch under which to save the notebook. Uses the repo default if not specified.
        commit_message: Message of the commit with the notebook upload. Default is ``"Uploaded notebook {name}"``
        versioning: Either ``"git"`` or ``"dvc"``.
        colab_timeout: For Colab environments sets the timeout for getting the notebook (in seconds).
            Raise this if you have a large notebook and encountering timeouts while saving it.


    .. note::
        Right now correctly saves only notebooks in a Colab environment.
        Regular Jupyter environment will have the execution history saved instead of the notebook.
    """

    if not _inside_notebook():
        log_message(
            "Trying to save a notebook while not being in an IPython environment. No notebook will be saved", logger
        )
        return

    # Handle file path
    file_path = Path(path)
    if file_path.name != "." and "." not in file_path.name:
        file_path /= _default_notebook_name()
    file_path = "/" / file_path
    remote_path = PurePosixPath("/") / file_path.as_posix()

    # Handle commit message
    if commit_message is None:
        commit_message = f"Uploaded notebook {file_path.name}"

    # Handle repo name
    parsed_repo = repo.split("/")
    if len(parsed_repo) != 2:
        raise RuntimeError(f'Repo format has to be "user/repo" (got {repo})')
    owner, repo = parsed_repo

    # Upload notebook
    with tempfile.TemporaryDirectory() as tmp:
        out_path = f"{tmp}/{file_path.name}"
        if _inside_colab():
            from google.colab import _message  # If inside colab, this import is guaranteed

            notebook_ipynb = _message.blocking_request("get_ipynb", timeout_sec=colab_timeout)
            if notebook_ipynb is None or "ipynb" not in notebook_ipynb:
                raise RuntimeError("Couldn't get notebook data from colab.")
            with open(out_path, "w") as file:
                file.write(json.dumps(notebook_ipynb["ipynb"], indent=4))
        else:
            log_message("Saving only the execution history for the notebook in Jupyter environments", logger)
            IPy.get_ipython().run_line_magic("notebook", out_path)

        repo = Repo(owner, repo, branch=branch)
        repo.upload(
            out_path,
            remote_path=remote_path.as_posix(),
            commit_message=commit_message,
            versioning=versioning,
            force=True,
        )



import datetime
import enum
import logging
from dataclasses import dataclass, field
from typing import Any, List, Union, Optional, Set

from dataclasses_json import config, DataClassJsonMixin
from dagshub.data_engine.dtypes import MetadataFieldType, ReservedTags

logger = logging.getLogger(__name__)


@dataclass
class Metadata:
    key: str
    value: Any


autogenerated_columns = {
    "path",
    "datapoint_id",
    "dagshub_download_url",
}


class IntegrationStatus(enum.Enum):
    VALID = "VALID"
    INVALID_CREDENTIALS = "INVALID_CREDENTIALS"
    MISSING = "MISSING"


class PreprocessingStatus(enum.Enum):
    READY = "READY"
    IN_PROGRESS = "IN_PROGRESS"
    AUTO_SCAN_IN_PROGRESS = "AUTO_SCAN_IN_PROGRESS"
    FAILED = "FAILED"


class DatasourceType(enum.Enum):
    BUCKET = "BUCKET"
    REPOSITORY = "REPOSITORY"
    CUSTOM = "CUSTOM"


class ScanOption(str, enum.Enum):
    """
    Enum of options that can be applied during scanning process with
    :func:`~dagshub.data_engine.model.datasource.Datasource.scan_source`
    """

    FORCE_REGENERATE_AUTO_SCAN_VALUES = "FORCE_REGENERATE_AUTO_SCAN_VALUES"
    """
    Regenerate all the autogenerated metadata values for the whole datasource
    """


@dataclass
class MetadataFieldSchema(DataClassJsonMixin):
    # This should match the GraphQL schema: MetadataFieldProps
    name: str
    valueType: MetadataFieldType = field(metadata=config(encoder=lambda val: val.value))
    multiple: bool
    tags: Optional[Set[str]]

    def __repr__(self):
        res = f"{self.name} ({self.valueType.value})"
        if self.tags is not None and len(self.tags) > 0:
            res += f" with tags: {self.tags}"
        return res

    def is_annotation(self):
        return ReservedTags.ANNOTATION.value in self.tags if self.tags else False

    def is_document(self):
        return ReservedTags.DOCUMENT.value in self.tags if self.tags else False

    def is_thumbnail(self):
        return ReservedTags.THUMBNAIL_VIZ.value in self.tags if self.tags else False


@dataclass
class MetadataSelectFieldSchema(MetadataFieldSchema):
    """
    This is used as fields in QueryResult and has additional fields over the base class
    """

    originalName: str
    autoGenerated: bool
    asOf: Optional[int]

    @staticmethod
    def from_metadata_field_schema(mfs: MetadataFieldSchema) -> "MetadataSelectFieldSchema":
        return MetadataSelectFieldSchema(
            name=mfs.name,
            originalName=mfs.name,
            autoGenerated=False,
            valueType=mfs.valueType,
            multiple=mfs.multiple,
            tags=mfs.tags,
            asOf=None,
        )


@dataclass
class DatasourceResult:
    id: Union[str, int]
    name: str
    rootUrl: str
    integrationStatus: IntegrationStatus
    preprocessingStatus: PreprocessingStatus
    type: DatasourceType
    metadataFields: Optional[List[MetadataFieldSchema]]


@dataclass
class DatasetResult:
    id: Union[str, int]
    name: str
    datasource: DatasourceResult
    datasetQuery: str


@dataclass
class DatapointHistoryResult:
    timestamp: datetime.datetime
    """Times of the version changes for this datapoint. The timezone is always UTC."""



import datetime
import enum
import logging
from typing import Optional, Union, Dict

import pytz
from treelib import Tree, Node

from dagshub.data_engine.model.errors import WrongOperatorError
from dagshub.data_engine.model.schema_util import metadata_type_lookup, metadata_type_lookup_reverse
from dagshub.data_engine.dtypes import MetadataFieldType

logger = logging.getLogger(__name__)


def bytes_deserializer(val: str) -> bytes:
    if val.startswith('b"') or val.startswith("b'"):
        return val[2:-1].encode()
    # Fallback - encode whatever we got from the server
    return val.encode()


_metadataTypeCustomConverters = {
    bool: lambda x: x.lower() == "true",
    bytes: bytes_deserializer,
    datetime.datetime: lambda x: datetime.datetime.fromtimestamp(int(x) / 1000).astimezone(pytz.utc),
}


class FieldFilterOperand(enum.Enum):
    EQUAL = "EQUAL"
    GREATER_THAN = "GREATER_THAN"
    GREATER_EQUAL_THAN = "GREATER_EQUAL_THAN"
    LESS_THAN = "LESS_THAN"
    LESS_EQUAL_THAN = "LESS_EQUAL_THAN"
    CONTAINS = "CONTAINS"
    IS_NULL = "IS_NULL"
    STARTS_WITH = "STARTS_WITH"
    ENDS_WITH = "ENDS_WITH"
    DATE_TIME_FILTER = "DATE_TIME_FILTER"


class FieldFilterDateTimeFilter(enum.Enum):
    YEAR = "YEAR"
    MONTH = "MONTH"
    DAY = "DAY"
    TIMEOFDAY = "TIMEOFDAY"


dt_range_ops = [
    FieldFilterDateTimeFilter.YEAR,
    FieldFilterDateTimeFilter.MONTH,
    FieldFilterDateTimeFilter.DAY,
    FieldFilterDateTimeFilter.TIMEOFDAY,
]


fieldFilterOperandMap = {
    "eq": FieldFilterOperand.EQUAL,
    "gt": FieldFilterOperand.GREATER_THAN,
    "ge": FieldFilterOperand.GREATER_EQUAL_THAN,
    "lt": FieldFilterOperand.LESS_THAN,
    "le": FieldFilterOperand.LESS_EQUAL_THAN,
    "contains": FieldFilterOperand.CONTAINS,
    "isnull": FieldFilterOperand.IS_NULL,
    "startswith": FieldFilterOperand.STARTS_WITH,
    "endswith": FieldFilterOperand.ENDS_WITH,
    "date_time_filter": FieldFilterOperand.DATE_TIME_FILTER,
}

fieldFilterDateTimeFilterMap = {
    "year": FieldFilterDateTimeFilter.YEAR,
    "month": FieldFilterDateTimeFilter.MONTH,
    "day": FieldFilterDateTimeFilter.DAY,
    "timeofday": FieldFilterDateTimeFilter.TIMEOFDAY,
}
fieldFilterOperandMapReverseMap: Dict[str, str] = {}

for k, v in fieldFilterOperandMap.items():
    fieldFilterOperandMapReverseMap[v.value] = k

for k, v in fieldFilterDateTimeFilterMap.items():
    fieldFilterOperandMapReverseMap[v.value] = k

UNFILLED_NODE_TAG = "undefined"


class QueryFilterTree:
    def __init__(
        self,
        column_or_query: Optional[Union[str, "QueryFilterTree"]] = None,
        field_as_of: Optional[int] = None,
    ):
        self._operand_tree: Tree = Tree()

        if type(column_or_query) is str:
            # If it's ds["column"] then the root node is just the column name, will be filled later
            data: Dict[str, Union[str, int]] = {"field": column_or_query}
            if field_as_of is not None:
                data["as_of"] = int(field_as_of)
            self._operand_tree.create_node(UNFILLED_NODE_TAG, data=data)
        elif column_or_query is not None:
            self._operand_tree.create_node(column_or_query)

    def __repr__(self):
        if self.is_empty:
            return "Query: empty"
        return f"Query: {self.tree_to_dict()}"

    @property
    def column_filter(self) -> Optional[str]:
        filter_node = self._column_filter_node
        if filter_node is None:
            return None
        return filter_node.data["field"]

    def compose(self, op: str, other: Optional[Union[str, int, float, "QueryFilterTree", datetime.datetime]]):
        """
        Compose the current query with another query or a value using the specified operator.

        Args:
            op (str): The operator to use for composing the query.
            other (Optional[Union[str, int, float, "QueryFilterTree", datetime.datetime]]):
                The query or value to compose with.

        Raises:
            RuntimeError: If the operation is not supported or if there is a mismatch in usage.

        Notes:
            - If the current query contains only a column filter,
                it will be composed into a tree with the specified operator and value.
            - If the operation is 'isnull', it can only be applied to a column filter;
                otherwise, a RuntimeError is raised.
            - If the operation is 'not', a 'not' node is added to the query tree.
            - If either query is empty, the composition is adjusted accordingly.

        Example:
            ```
            ds['col1'] > 5
            ds['col2'].is_null()
            ```
            The above queries can be composed as:
            ```
            (ds['col1'] > 5) & (ds['col2'].is_null())
            ```
        """
        if self._column_filter_node is not None:
            # If there was an unfilled query node with a column - put the operand in that node
            node = self._column_filter_node
            if op.startswith("!"):
                # Negation in the operation - prepend a not node before the current node
                tree = self._operand_tree
                parent_id = tree.parent(node.identifier)
                if parent_id is None:
                    # Root node - need to recreate a new tree with the not node as the top node
                    new_tree = Tree()
                    not_node = new_tree.create_node("not")
                    new_tree.paste(not_node.identifier, self._operand_tree)
                    self._operand_tree = new_tree
                else:
                    not_node = tree.create_node("not", parent=parent_id)
                    tree.move_node(node.identifier, not_node.identifier)
                op = op[1:]
            node.tag = op
            node.data.update({"value": other})
        elif op == "isnull":
            # Can only do isnull on the column filter, if we got here, there's something wrong
            raise RuntimeError("is_null operation can only be done on a column (e.g. ds['col1'].is_null())")
        elif op == "not":
            new_tree = Tree()
            not_node = new_tree.create_node("not")
            new_tree.paste(not_node.identifier, self._operand_tree)
            self._operand_tree = new_tree
        else:
            # The other side is an actual query with its own tree - make a subtree
            if type(other) is not QueryFilterTree:
                raise RuntimeError(f"Expected other argument to be a dataset, got {type(other)} instead")
            if op not in ["and", "or"]:
                raise RuntimeError(
                    f"Cannot use operator '{op}' to chain two queries together.\r\n"
                    f"Queries:\r\n"
                    f"\t{self}\r\n"
                    f"\t{other}\r\n"
                )
            # Don't compose with an empty query, carry the other instead
            if self.is_empty:
                self._operand_tree = other._operand_tree
                return
            elif other.is_empty and other._column_filter_node is None:
                return
            composite_tree = Tree()
            root_node = composite_tree.create_node(op)
            composite_tree.paste(root_node.identifier, self._operand_tree)
            composite_tree.paste(root_node.identifier, other._operand_tree)
            self._operand_tree = composite_tree

    @property
    def _column_filter_node(self) -> Node:
        return next(self._operand_tree.filter_nodes(lambda n: n.tag == UNFILLED_NODE_TAG), None)

    @property
    def _operand_root(self) -> Node:
        return self._operand_tree[self._operand_tree.root]

    def serialize(self) -> Optional[Dict]:
        if self.is_empty:
            return None
        return self._serialize_node(self._operand_root, self._operand_tree)

    @staticmethod
    def _serialize_node(node: Node, tree: Tree) -> Dict:
        operand = node.tag
        if operand in ["and", "or"]:
            # recursively serialize children subqueries
            return {operand: [QueryFilterTree._serialize_node(child, tree) for child in tree.children(node.identifier)]}
        if operand == "not":
            assert len(tree.children(node.identifier)) == 1
            child = tree.children(node.identifier)[0]
            serialized = QueryFilterTree._serialize_node(child, tree)
            serialized["not"] = True
            return serialized
        else:
            # op can be a simple comparator, or a type of timeFilter
            query_op = fieldFilterOperandMap.get(operand) or fieldFilterDateTimeFilterMap.get(operand)
            if query_op is None:
                raise WrongOperatorError(f"Operator {operand} is not supported")
            key = node.data["field"]
            value = node.data["value"]
            as_of = node.data.get("as_of")

            value_type = metadata_type_lookup[type(value)].value if type(value) in metadata_type_lookup else None

            # if one of basic value types:
            if value_type:
                if type(value) is bytes:
                    # TODO: this will need to probably be changed when we allow actual binary field comparisons
                    value = value.decode("utf-8")
                else:
                    if isinstance(value, datetime.datetime):
                        value = int(value.timestamp() * 1000)
                    else:
                        value = str(value)

            if value_type is None and query_op not in dt_range_ops:
                raise RuntimeError(
                    f"Value type {value_type} is not supported for querying.\r\n"
                    f"Supported types: {list(metadata_type_lookup.keys())}"
                )
            res = {
                "filter": {
                    "key": key,
                    "value": str(value),
                    "valueType": value_type,
                    "comparator": query_op.value,
                }
            }

            if as_of:
                res["filter"]["asOf"] = as_of

            if query_op in dt_range_ops:
                # value is the actual node value in timeofday ("HH:mm-HH:mm"),
                # else, we use valueRange, so unset value
                res["filter"]["value"] = value if query_op is FieldFilterDateTimeFilter.TIMEOFDAY else 0
                # value type from client perspective is string or list,
                # but we need to tell backend to get datetime from db
                res["filter"]["valueType"] = MetadataFieldType.DATETIME.value
                # we use valueRange unless the node op is timeofday,
                # else, we use value, so unset valueRange
                res["filter"]["valueRange"] = 0 if query_op is FieldFilterDateTimeFilter.TIMEOFDAY else value
                # timeFilter replaces comparator
                res["filter"]["timeFilter"] = query_op.value
                # comparator indicates that timeFilter replaces comparator
                res["filter"]["comparator"] = FieldFilterOperand.DATE_TIME_FILTER.value

            return res

    @staticmethod
    def deserialize(serialized_query: Union["QueryFilterTree", Dict]) -> "QueryFilterTree":
        if isinstance(serialized_query, QueryFilterTree):
            return serialized_query
        q = QueryFilterTree()
        op_tree = Tree()
        QueryFilterTree._deserialize_node(serialized_query, op_tree)
        q._operand_tree = op_tree
        return q

    @staticmethod
    def _deserialize_node(node_dict: Dict, tree: Tree, parent_node=None) -> None:
        keys = list(node_dict.keys())
        if len(keys) == 0:
            return

        is_negative = node_dict.get("not", False)
        if is_negative:
            # If operation is negative - prepend a "not" node to the node we'll be adding
            neg_node = Node(tag="not")
            tree.add_node(neg_node, parent_node)
            parent_node = neg_node

        op_type = keys[0]
        val = node_dict[op_type]
        # Types: and, or, filter
        if op_type == "filter":
            comparator = fieldFilterOperandMapReverseMap[val["comparator"]]
            key = val["key"]
            if comparator.upper() == FieldFilterOperand.DATE_TIME_FILTER.value:
                # if comparator indicates that timeFilter replaced comparator,
                # we need to rebuild node carefully

                # value type must ignore actual value and be later set to MetadataFieldType.DATETIME
                value_type = None

                if val["timeFilter"] == FieldFilterDateTimeFilter.TIMEOFDAY.value:
                    value = val["value"]
                else:
                    value = val["valueRange"]

                # timeFilter replaced comparator in query, so now the reverse action
                comparator = val["timeFilter"].lower()
            else:
                value_type = metadata_type_lookup_reverse[val["valueType"]]
                converter = _metadataTypeCustomConverters.get(value_type, lambda x: value_type(x))
                value = converter(val["value"])
            as_of = val.get("asOf")
            node = Node(tag=comparator, data={"field": key, "value": value, "as_of": as_of})
            tree.add_node(node, parent_node)
        elif op_type in ("and", "or"):
            main_node = Node(tag=op_type)
            tree.add_node(main_node, parent_node)
            for nested_node in val:
                QueryFilterTree._deserialize_node(nested_node, tree, main_node)
        else:
            raise RuntimeError(f"Unknown serialized query dict: {node_dict}")

    def tree_to_dict(self):
        return self._operand_tree.to_dict(with_data=True)

    def __deepcopy__(self, memodict={}):
        q = QueryFilterTree()
        q._operand_tree = Tree(tree=self._operand_tree, deep=True)
        return q

    @property
    def is_empty(self):
        return self._operand_tree.root is None or self._column_filter_node is not None



from dataclasses import dataclass, field
from typing import Set, Dict, Optional

from dataclasses_json import DataClassJsonMixin, config, LetterCase

from dagshub.common.util import exclude_if_none
from dagshub.data_engine.dtypes import MetadataFieldType


@dataclass
class DatasourceFieldInfo:
    """
    Utility class for add_metadata that caches a bunch of relevant information about the datasource fields,
    to prevent constant lookups
    """

    multivalue_fields: Set[str]
    field_value_types: Dict[str, MetadataFieldType]
    document_fields: Set[str]


@dataclass
class DatapointMetadataUpdateEntry(DataClassJsonMixin):
    url: str
    key: str
    value: str
    valueType: MetadataFieldType = field(metadata=config(encoder=lambda val: val.value))
    allowMultiple: bool = False
    timeZone: Optional[str] = field(
        default=None, metadata=config(exclude=exclude_if_none, letter_case=LetterCase.CAMEL)
    )



import logging
import os
from dataclasses import dataclass
from enum import auto, IntEnum
from functools import cached_property
from pathlib import Path, PurePosixPath
from typing import Optional, Literal, Union, Tuple, List

import dacite
import yaml
from pathvalidate import sanitize_filepath

from dagshub.common.analytics import send_analytics_event
from dagshub.common.api import RepoAPI
from dagshub.common.api.repo import PathNotFoundError
from dagshub.common.api.responses import StorageAPIEntry
from dagshub.common.determine_repo import determine_repo
from dagshub.common.helpers import log_message
from dagshub.common.util import is_path_relative_to
from dagshub.models.model_loaders import (
    ModelLoader,
    RepoModelLoader,
    BucketModelLoader,
    DagsHubStorageModelLoader,
)

# TODO: don't forget to separate dvc downloads by the hash of the dir

logger = logging.getLogger(__name__)


class ModelNotFoundError(Exception):
    def __str__(self):
        return "Could not find model"


@dataclass
class ModelYamlStruct:
    model_dir: str


class StorageType(IntEnum):
    Repo = auto()
    DagshubStorage = auto()
    Bucket = auto()


class ModelLocator:
    def __init__(
        self,
        repo: Optional[Union[str, RepoAPI]] = None,
        path: Optional[str] = None,
        bucket: Optional[str] = None,
        git_ref: Optional[str] = None,
        mlflow_model: Optional[str] = None,
        mlflow_artifact: Optional[str] = None,
        download_dest: Optional[Union[str, os.PathLike]] = None,
        download_type: Literal["lazy", "eager"] = "eager",
    ):
        self._repo = repo
        self.path = path
        self.bucket = bucket
        self._git_ref = git_ref
        self.mlflow_model = mlflow_model
        self.mlflow_artifact = mlflow_artifact
        self._download_dest = download_dest
        self.download_type = download_type

    @cached_property
    def repo_api(self) -> RepoAPI:
        if isinstance(self._repo, RepoAPI):
            return self._repo
        if self._repo is None:
            repo_api, git_ref = determine_repo()
            if self._git_ref is None:
                self._git_ref = git_ref
            return repo_api
        return RepoAPI(self._repo)

    @property
    def git_ref(self):
        if self._git_ref is None:
            self._git_ref = self.repo_api.default_branch
        return self._git_ref

    @cached_property
    def download_destination(self) -> Path:
        if self._download_dest is not None:
            return Path(self._download_dest)
        return Path(sanitize_filepath(os.path.join(Path.home(), "dagshub", "models", self.repo_api.full_name)))

    @cached_property
    def repo_storages(self) -> List[StorageAPIEntry]:
        return self.repo_api.get_connected_storages()

    def _handle_path(self, path: str) -> Tuple[str, StorageType]:
        """
        Handles user-printed path, determining where it belongs (repo, bucket or repo's bucket).

        If path is a repo path, returns (path as is, StorageType.Repo)
        If path is a storage path, ({actual path in storage} + StorageType.Bucket/DagshubStorage)
        Storage path resolution is based on bucket name, if the path starts with a name of an existing bucket,
        then greedily return that bucket as the bucket with the path.
        """
        str_path = str(path)

        if "/" not in str_path:
            return path, StorageType.Repo

        if str_path.startswith("dagshub_storage/"):
            return str_path[len("dagshub_storage/") :], StorageType.DagshubStorage
        in_path = PurePosixPath(str_path)
        for storage in self.repo_storages:
            storage_path = PurePosixPath(storage.name)
            if is_path_relative_to(in_path, storage_path):
                bucketPath = f"{storage.protocol}/{str_path}"
                return bucketPath, StorageType.Bucket
        return path, StorageType.Repo

    def try_load_yaml(self, yaml_path=".dagshub/model.yaml") -> Optional[ModelLoader]:
        """
        Loads the info about where the model stored from a .yaml file
        """
        try:
            yaml_contents = self.repo_api.get_file(yaml_path)
        except PathNotFoundError:
            return None
        cfg = dacite.from_dict(ModelYamlStruct, yaml.safe_load(yaml_contents))

        model_path, storage_type = self._handle_path(cfg.model_dir)
        log_message(f"Loading the model from yaml file {yaml_path}", logger=logger)
        if storage_type == StorageType.DagshubStorage:
            try:
                self.repo_api.list_storage_path(f"s3/{self.repo_api.repo_name}/{model_path}")
                log_message(
                    f"Loading the model from DagsHub Storage {model_path}",
                    logger=logger,
                )
                return DagsHubStorageModelLoader(self.repo_api, PurePosixPath(model_path))
            except PathNotFoundError:
                raise ModelNotFoundError
        elif storage_type == StorageType.Bucket:
            try:
                self.repo_api.list_storage_path(model_path)
                log_message(f"Loading the model from bucket {model_path}", logger=logger)
                return BucketModelLoader(self.repo_api, PurePosixPath(model_path))
            except PathNotFoundError:
                raise ModelNotFoundError
        else:
            try:
                self.repo_api.list_path(model_path, revision=self.git_ref)
                log_message(f"Loading the model from repo {model_path}", logger=logger)
                return RepoModelLoader(self.repo_api, self.git_ref, PurePosixPath(model_path))
            except PathNotFoundError:
                raise ModelNotFoundError

    lookup_dirs = ["model", "models"]

    def try_find_model_in_repo(self) -> Optional[ModelLoader]:
        for lookup_dir in self.lookup_dirs:
            modelPath = PurePosixPath(lookup_dir)
            try:
                self.repo_api.list_path(str(modelPath), revision=self.git_ref)
                log_message(f"Loading the model from repo directory {modelPath}", logger=logger)
                return RepoModelLoader(repo_api=self.repo_api, revision=self.git_ref, path=modelPath)
            except PathNotFoundError:
                continue
        return None

    def try_find_model_in_bucket(self, bucket: str) -> Optional[ModelLoader]:
        bucketPath: Optional[PurePosixPath] = None
        for storage in self.repo_storages:
            if storage.name.startswith(bucket):
                bucketPath = PurePosixPath(f"{storage.protocol}/{storage.name}")
                break

        if bucketPath is None:
            return None

        for lookup_dir in self.lookup_dirs:
            modelPath = bucketPath / lookup_dir
            try:
                self.repo_api.list_storage_path(str(modelPath))
                log_message(
                    f"Loading the model from bucket directory {modelPath}",
                    logger=logger,
                )
                return BucketModelLoader(repo_api=self.repo_api, path=modelPath)
            except PathNotFoundError:
                continue

        return None

    def try_find_model_in_dagshub_storage(self) -> Optional[ModelLoader]:
        bucketPath = PurePosixPath("s3") / self.repo_api.repo_name
        for modelPath in self.lookup_dirs:
            try:
                self.repo_api.list_storage_path(str(bucketPath / modelPath))
                log_message(
                    f"Loading the model from DagsHub Storage directory {modelPath}",
                    logger=logger,
                )
                return DagsHubStorageModelLoader(repo_api=self.repo_api, path=PurePosixPath(modelPath))
            except PathNotFoundError:
                continue

        return None

    def find_model(self) -> ModelLoader:
        model_loader: Optional[ModelLoader]

        if self.path is not None:
            # Check that the dir exists and return that
            handled_path, storage_type = self._handle_path(self.path)
            try:
                if storage_type == storage_type.Bucket:
                    self.repo_api.list_storage_path(handled_path)
                    return BucketModelLoader(self.repo_api, PurePosixPath(handled_path))
                elif storage_type == storage_type.DagshubStorage:
                    self.repo_api.list_storage_path(PurePosixPath("s3") / self.repo_api.repo_name / handled_path)
                    return DagsHubStorageModelLoader(self.repo_api, PurePosixPath(handled_path))
                else:
                    self.repo_api.list_path(handled_path, self.git_ref)
                    return RepoModelLoader(self.repo_api, self.git_ref, PurePosixPath(handled_path))
            except PathNotFoundError:
                raise ModelNotFoundError

        if self.bucket is not None:
            model_loader = self.try_find_model_in_bucket(self.bucket)
            if model_loader is None:
                raise ModelNotFoundError
            if model_loader is not None:
                return model_loader

        # TODO: add mlflow here

        model_loader = self.try_load_yaml()
        if model_loader is not None:
            return model_loader

        model_loader = self.try_find_model_in_repo()
        if model_loader is not None:
            return model_loader

        model_loader = self.try_find_model_in_dagshub_storage()
        if model_loader is not None:
            return model_loader

        raise ModelNotFoundError

    def get_model_path(self) -> Path:
        send_analytics_event("Client_Models_GetModelPath")
        model_loader = self.find_model()
        return model_loader.load_model(self.download_type, self.download_destination, self.git_ref)


def get_model_path(
    repo: Optional[str] = None,
    path: Optional[str] = None,
    bucket: Optional[str] = None,
    git_ref: Optional[str] = None,
    # mlflow_model: Optional[str] = None,
    # mlflow_artifact: Optional[str] = None,
    download_dest: Optional[Union[str, os.PathLike]] = None,
    download_type: Literal["lazy", "eager"] = "eager",
) -> Path:
    """
    Load a model path from a DagsHub repository in way that is compatible with the Hugging Face Transformers library.

    Example usage:

    .. code-block:: python

       from transformers import AutoModel
       from dagshub.models import get_model_path

       model = AutoModel.from_pretrained(get_model_path("<user_name>/<repo_name>"))

    The function looks for the model in following places in the repository, in this order:

    - **.dagshub/model.yaml** file in the repository. The file format should be::

        model_dir: <path_to_model_dir_in_repo>

    - **model** and **models** dvc/git directories in the root of the repository
    - **model** and **models** directories in repository's DagsHub Storage

    If either ``path`` or ``bucket`` argument is specified, this lookup gets ignored.

    After a model has been found, it is either downloaded or mounted to a local directory,
    and the path to the directory with the model files is returned.

    Args:
        repo: Name of the repository to load the model from in the format of ``<user>/<repo>``. \
        If ``None`` tries to get a DagsHub repository from a git repository in the current working directory
        path: Path to the directory with the model in the repository. This skips the model resolution step.
        bucket: Name of the bucket with the model. If specified, the model will be loaded from either \
        ``<bucket>/model`` or ``<bucket>/models`` directories. This skips the model resolution step.
        git_ref: Specific git revision/branch to load model from. This only works for models versioned with dvc.
        download_dest: Path to the directory where to download the model. If ``None``, the model will be downloaded to \
        ``~/dagshub/models/<user>/<repo>/<path_to_model_in_repo>``. If specified, the model will be downloaded to \
        ``<download_dest>/<path_to_model_in_repo>``
        download_type: How to load the models. Possible values:

            - ``lazy``: Mounts a streaming filesystem with :func:`install_hooks() <dagshub.streaming.install_hooks>` \
            that will download files on demand.
            - ``eager``: Downloads the whole model directory at once. Faster for models with a lot of files.

    .. warning::
        In order for download to work consistently between eager and lazy loading, the resulting path to the model \
        is always included in the returned path.

    .. note::
        If using .dagshub/model.yaml file or a path argument, and you want to use a bucket, specify:

        - ``dagshub_storage/<path_to_model_dir>`` to point to a folder in DagsHub Storage
        - ``<bucket_name>/<path_to_model_dir>`` if using an integrated bucket. For ``s3://my-bucket``,\
         the bucket name is ``my-bucket``

    Returns:
        Path to the directory with the models.
    """
    loader = ModelLocator(
        repo=repo,
        path=path,
        bucket=bucket,
        git_ref=git_ref,
        # mlflow_model=mlflow_model,
        # mlflow_artifact=mlflow_artifact,
        download_dest=download_dest,
        download_type=download_type,
    )
    return loader.get_model_path()



import rich.progress

from dagshub.common import rich_console, config


def get_rich_progress(*additional_columns, transient=True):
    progress = rich.progress.Progress(
        rich.progress.SpinnerColumn(),
        *rich.progress.Progress.get_default_columns(),
        *additional_columns,
        console=rich_console,
        transient=transient,
        disable=config.quiet
    )
    return progress



from .transforms import run_preupload_transforms
from .validation import (
    validate_uploading_metadata,
    precalculate_metadata_info,
    MultipleDataTypesUploadedError,
    StringFieldValueTooLongError,
)

__all__ = [
    run_preupload_transforms.__name__,
    validate_uploading_metadata.__name__,
    precalculate_metadata_info.__name__,
    MultipleDataTypesUploadedError.__name__,
    StringFieldValueTooLongError.__name__,
]
